/*
** Copyright (C) 2003 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include <glib.h>
#include "dialogprintproject.h"
#include "utilities.h"
#include "bible.h"
#include "book.h"
#include "chapter.h"
#include "usfm.h"
#include "usfmtools.h"
#include "log.h"
#include "html_book.h"
#include "htmlbrowser.h"
#include <config.h>
#include "html_utils.h"
#include <time.h>
#include <set>


PrintProjectDialog::PrintProjectDialog (Scripture *scripture_in, const string& projectname)
{
  // Store data.
  scripture = scripture_in;
  project_name = projectname;

  dialogprintproject = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (dialogprintproject), "Print project");
  gtk_window_set_position (GTK_WINDOW (dialogprintproject), GTK_WIN_POS_CENTER_ON_PARENT);
  gtk_window_set_modal (GTK_WINDOW (dialogprintproject), TRUE);
  gtk_window_set_type_hint (GTK_WINDOW (dialogprintproject), GDK_WINDOW_TYPE_HINT_DIALOG);

  dialog_vbox1 = GTK_DIALOG (dialogprintproject)->vbox;
  gtk_widget_show (dialog_vbox1);

  vbox1 = gtk_vbox_new (FALSE, 5);
  gtk_widget_show (vbox1);
  gtk_box_pack_start (GTK_BOX (dialog_vbox1), vbox1, TRUE, TRUE, 0);

  label1 = gtk_label_new ("Printing project, please wait...");
  gtk_widget_show (label1);
  gtk_box_pack_start (GTK_BOX (vbox1), label1, FALSE, FALSE, 0);

  progressbar1 = gtk_progress_bar_new ();
  gtk_widget_show (progressbar1);
  gtk_box_pack_start (GTK_BOX (vbox1), progressbar1, FALSE, FALSE, 0);

  dialog_action_area1 = GTK_DIALOG (dialogprintproject)->action_area;
  gtk_widget_show (dialog_action_area1);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area1), GTK_BUTTONBOX_END);

  g_signal_connect (G_OBJECT (dialogprintproject), "delete_event", G_CALLBACK (delete_event), gpointer(this));

  // Start updating the progress bar.
  percentage = 0;
  event_source_id1 = g_timeout_add_full (G_PRIORITY_DEFAULT, 50, GSourceFunc (on_update_display_timeout), gpointer(this), NULL);
  // Start the conversion process.
  dialog_is_about_to_be_deleted = false;
  g_thread_create (GThreadFunc (printreferencesdialog_convert), gpointer(this), false, NULL);
}


PrintProjectDialog::~PrintProjectDialog ()
{
  gtk_widget_destroy (dialogprintproject);
}


int
PrintProjectDialog::run ()
{
  return gtk_dialog_run (GTK_DIALOG (dialogprintproject));
}


gboolean PrintProjectDialog::delete_event(GtkWidget *widget, GdkEvent  *event, gpointer   data )
{
  // Call handler.
  ((PrintProjectDialog *) data)->on_delete_event ();
  // Return true to destroy the window.
  return true;
}


void PrintProjectDialog::on_delete_event ()
{
  // Set flag we're about to get destroyed.
  dialog_is_about_to_be_deleted = true;
  // Destroy the two event sources, so that the timeout functions are no longer called.
  GSource *source;
  source = g_main_context_find_source_by_id (NULL, event_source_id1);
  if (source)
    g_source_destroy (source);
  source = g_main_context_find_source_by_id (NULL, event_source_id2);
  if (source)
    g_source_destroy (source);
  // Wait a while to let functions finish their work.
  g_usleep (40000);
}


bool PrintProjectDialog::on_update_display_timeout (gpointer data)
{
  return ((PrintProjectDialog*) data)->update_display();
}


bool PrintProjectDialog::update_display ()
{
  // Update progressbar.
  gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (progressbar1), percentage);
  // If the progressbar is full then stop updating the display.
  if (percentage == 1) {
    return false;
  } else {
    return true;
  }
}



void PrintProjectDialog::printreferencesdialog_convert (gpointer data)
{
  ((PrintProjectDialog *) data)->convert ();
}


void PrintProjectDialog::convert ()
{
  log_message ("Starting to print the project");

  // Create directory to put everything in.
  string working_directory = g_build_filename(g_get_tmp_dir(), "bibledit_project", NULL);
  create_directory (working_directory);

  /*
    We'll create one index html file
    and one html file for each book.
    Navigators are added to each html file, both at the top and the bottom.
    To do this properly we first get the meta-data for each html file.
    This includes its title, link to previous file and link to next file,
    and for the index file all links to the books.
  */

  // Storage for meta data.
  vector<unsigned int> scriptureindexes; // Indexes in object Scripture.
  vector<string> booktitles;             // Titles of the books.
  vector<string> booklinks;              // Links of the books.
  vector<string> previoustitles;         // Title for the "Previous" link.
  vector<string> previouslinks;          // Filename for the "Previous" link
  vector<string> nexttitles;             // Title for the "Next" link.
  vector<string> nextlinks;              // Filename for the "Next" link

  // Go through all the books and collect the metadata for each book.
  for (unsigned int i = 0; i < scripture->books.size(); i++) {
    // Dig up the metadata.
    unsigned int scriptureindex = i;
    
    // Get the title from the book. The title will be the running header.
    string booktitle;
    Book book (scripture->paths[i], false, false);
    // Get the text of the first chapter.
    vector <ustring> lines;
    book.get_chapter (1, lines);
    // Search for data in the first couple of lines.
    unsigned int high_limit;
    high_limit = MIN (10, lines.size());
    Usfm usfm(0);
    for (unsigned int i2 = 0; i2 < high_limit; i2++) {
      ustring line = lines[i2];
      ustring marker = usfm_extract_marker (line);
      if (usfm.is_runningheader (marker))
        booktitle = line;
    }      
    // No running header found? Take English name.
    if (booktitle.length() == 0) 
      booktitle = scripture->books[i];
    // Get the name for the link.
    string booklink = booktitle + ".html";
    // Store metadata.
    scriptureindexes.push_back(scriptureindex);
    booktitles.push_back(booktitle);
    booklinks.push_back(booklink);
  }

  // Fill in data for "previous" and "next" links.
  for (unsigned int i = 0; i < scriptureindexes.size(); i++) {
    // Get data.
    string previoustitle;
    string previouslink;
    if (i > 0) {
      previoustitle = booktitles[i - 1];
      previouslink = previoustitle + ".html";
    } 
    string nexttitle;
    string nextlink;
    if (i < scriptureindexes.size() - 1) {
      nexttitle = booktitles[i + 1];
      nextlink = nexttitle + ".html";
    }
    // Store data.
    previoustitles.push_back(previoustitle);
    previouslinks.push_back(previouslink);
    nexttitles.push_back(nexttitle);
    nextlinks.push_back(nextlink);
  }
  
  // Update links for index file.
  if (scriptureindexes.size()) {
    previoustitles[0] = project_name;
    previouslinks[0] = "index.html";
  }
  
  // Produce the index file.  
  string htmlfilename = g_build_filename(working_directory.c_str(), "index.html", NULL);
  // Note: Object htmlbook should go out of scope to get the output file written.
  {
    // Get some variables for the navigation block.
    string nexttitle;
    string nextlink;
    if (booktitles.size()) {
      nexttitle = booktitles[0];
      nextlink = booklinks[0];
    }
    // Start producing the html file.
    vector<string> dummy;
    HtmlBook htmlbook (htmlfilename, project_name, "", "", nexttitle, nextlink, "index.html", dummy);
    htmlbook.openindex();
    // Write links to all the books.
    for (unsigned int i = 0; i < booktitles.size(); i++) {
      htmlbook.writelink (booktitles[i], booklinks[i]);
    }
    htmlbook.closeindex();
  }

  // View the index file. (Other files still have to be generated.)
  htmlbrowser (htmlfilename);

  // Go through the metadata and generate all books.
  for (unsigned int i = 0; i < scriptureindexes.size(); i++) {
    // Close window?
    if (dialog_is_about_to_be_deleted) return;
    // Calculate the percentage complete.
    percentage = double (i) / double (scripture->books.size());
    // Search for the headers(s) of the book, as printed above chapter one.    
    vector<string> headers;
    {
      Book book (scripture->paths[scriptureindexes[i]], false, false);
      // Get the text of the first chapter.
      vector <ustring> lines;
      book.get_chapter (1, lines);
      // Search for data in the first couple of lines.
      unsigned int high_limit;
      high_limit = MIN (10, lines.size());
      Usfm usfm(0);
      for (unsigned int i2 = 0; i2 < high_limit; i2++) {
        ustring line = lines[i2];
        ustring marker = usfm_extract_marker (line);
        if (usfm.is_maintitle (marker) || usfm.is_subtitle (marker))
          headers.push_back(lines[i2]);
      }      
    }
    // Filename to be written.
    htmlfilename = g_build_filename(working_directory.c_str(), booklinks[i].c_str(), NULL);
    HtmlBook htmlbook (htmlfilename, booktitles[i], previoustitles[i], previouslinks[i], nexttitles[i], nextlinks[i], "index.html", headers);
    // Open the book.
    Book book (scripture->paths[scriptureindexes[i]], false, false);
    // Go through all the chapters of the book.
    for (unsigned int i2 = 1; i2 <= book.number_of_chapters; i2++) {
      htmlbook.openchapter (i2, booktitles[i]);
      // Get the text of each chapter.
      vector <ustring> lines;
      book.get_chapter (i2, lines);
      htmlbook.openparagraph ();
      // Go through the lines and mark unwanted or unprintable ones.
      set<int> lines_to_remove;
      Usfm usfm(0);
      for (unsigned int i3 = 0; i3 < lines.size(); i3++) {
        // Get the marker.
        ustring line = lines[i3];
        ustring marker = usfm_extract_marker (line);
        // id - unprintable.
        if (usfm.is_id(marker))
          lines_to_remove.insert(i3);
        // chapter - already printed, so unwanted here.
        if (usfm.is_chapter(marker))
          lines_to_remove.insert(i3);
        // running header - already printed, so unwanted here.
        if (usfm.is_runningheader(marker))
          lines_to_remove.insert(i3);
        // Titles already printed -> unwanted here.
        if (usfm.is_maintitle(marker) || usfm.is_subtitle(marker))
          lines_to_remove.insert(i3);
      }
      // Go through each line.
      for (unsigned int i3 = 0; i3 < lines.size(); i3++) {
        // Skip unwanted/unprintable lines.
        if (lines_to_remove.find(i3) == lines_to_remove.end()) {
          vector <size_t> dummy;
          html_handle_entities (lines[i3], dummy);
          // Deal with the line here.
          ustring s (lines[i3]);
          ustring marker = usfm_extract_marker(s);
          if (usfm.is_sectionheading1(marker)) {
            htmlbook.openparagraph();
            htmlbook.writeline ("<b>" + s + "</b>");
            htmlbook.closeparagraph();
          } else if (usfm.is_paragraph(marker)) {
            htmlbook.openparagraph();
          } else {
            htmlbook.writeline (s);
          }
        }
      }
      htmlbook.closeparagraph ();
      htmlbook.closechapter ();
    }    
  }

  // Log: ready.
  log_message ("Ready printing the project");
  
  // We're now done.
  percentage = 1;
  gtk_label_set_text (GTK_LABEL (label1), "The browser has the project.");

  // Close window after a while.
  event_source_id2 = g_timeout_add_full (G_PRIORITY_DEFAULT, 5000, GSourceFunc (on_close_timeout), gpointer(this), NULL);
}


bool
PrintProjectDialog::on_close_timeout (gpointer data)
{
  return ((PrintProjectDialog *) data)->on_close ();
}


bool
PrintProjectDialog::on_close ()
{
  gtk_dialog_response (GTK_DIALOG (dialogprintproject), GTK_RESPONSE_OK);
  return false;
}
