/*
** Copyright (C) 2003 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include <glib.h>
#include "dialogprintreferences.h"
#include "utilities.h"
#include "bible.h"
#include "book.h"
#include "chapter.h"
#include "usfm.h"
#include "usfmtools.h"
#include "log.h"
#include "html_list.h"
#include "htmlbrowser.h"
#include <config.h>
#include "html_utils.h"


PrintReferencesDialog::PrintReferencesDialog (const string& projectname, const vector <string> &references_in, const vector <ustring>& highlights_in, const vector <bool>& casesensitives_in):
scripture (projectname)
{
  // Store data.
  references.assign (references_in.begin(), references_in.end());
  highlightwords.assign (highlights_in.begin(), highlights_in.end());
  casesensitives.assign (casesensitives_in.begin(), casesensitives_in.end());

  dialogprintreferences = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (dialogprintreferences), "Print references");
  gtk_window_set_position (GTK_WINDOW (dialogprintreferences), GTK_WIN_POS_CENTER_ON_PARENT);
  gtk_window_set_modal (GTK_WINDOW (dialogprintreferences), TRUE);
  gtk_window_set_type_hint (GTK_WINDOW (dialogprintreferences), GDK_WINDOW_TYPE_HINT_DIALOG);

  dialog_vbox1 = GTK_DIALOG (dialogprintreferences)->vbox;
  gtk_widget_show (dialog_vbox1);

  vbox1 = gtk_vbox_new (FALSE, 5);
  gtk_widget_show (vbox1);
  gtk_box_pack_start (GTK_BOX (dialog_vbox1), vbox1, TRUE, TRUE, 0);

  label1 = gtk_label_new ("Printing references, please wait...");
  gtk_widget_show (label1);
  gtk_box_pack_start (GTK_BOX (vbox1), label1, FALSE, FALSE, 0);

  progressbar1 = gtk_progress_bar_new ();
  gtk_widget_show (progressbar1);
  gtk_box_pack_start (GTK_BOX (vbox1), progressbar1, FALSE, FALSE, 0);

  dialog_action_area1 = GTK_DIALOG (dialogprintreferences)->action_area;
  gtk_widget_show (dialog_action_area1);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area1), GTK_BUTTONBOX_END);

  g_signal_connect (G_OBJECT (dialogprintreferences), "delete_event", G_CALLBACK (delete_event), gpointer(this));

  // Start updating the progress bar.
  percentage = 0;
  event_source_id1 = g_timeout_add_full (G_PRIORITY_DEFAULT, 50, GSourceFunc (on_update_display_timeout), gpointer(this), NULL);
  // Start the conversion process.
  dialog_is_about_to_be_deleted = false;
  g_thread_create (GThreadFunc (printreferencesdialog_convert), gpointer(this), false, NULL);
}


PrintReferencesDialog::~PrintReferencesDialog ()
{
  gtk_widget_destroy (dialogprintreferences);
}


int
PrintReferencesDialog::run ()
{
  return gtk_dialog_run (GTK_DIALOG (dialogprintreferences));
}


gboolean PrintReferencesDialog::delete_event(GtkWidget *widget, GdkEvent  *event, gpointer   data )
{
  // Call handler.
  ((PrintReferencesDialog *) data)->on_delete_event ();
  // Return true to destroy the window.
  return true;
}


void PrintReferencesDialog::on_delete_event ()
{
  // Set flag we're about to get destroyed.
  dialog_is_about_to_be_deleted = true;
  // Destroy the two event sources, so that the timeout functions are no longer called.
  GSource *source;
  source = g_main_context_find_source_by_id (NULL, event_source_id1);
  if (source)
    g_source_destroy (source);
  source = g_main_context_find_source_by_id (NULL, event_source_id2);
  if (source)
    g_source_destroy (source);
  // Wait a while to let functions finish their work.
  g_usleep (40000);
}


bool PrintReferencesDialog::on_update_display_timeout (gpointer data)
{
  return ((PrintReferencesDialog*) data)->update_display();
}


bool PrintReferencesDialog::update_display ()
{
  // Update progressbar.
  gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (progressbar1), percentage);
  // If the progressbar is full stop updating the display.
  if (percentage == 1) {
    return false;
  } else {
    return true;
  }
}



void PrintReferencesDialog::printreferencesdialog_convert (gpointer data)
{
  ((PrintReferencesDialog *) data)->convert ();
}


void PrintReferencesDialog::convert ()
{
  log_message ("Starting to print the references");

  // Produce filename of html file.
  string htmlfilename = g_build_filename(g_get_tmp_dir(), "bibledit_references.html", NULL);

  // Start producing the html file.
  // Note: Object HtmlList should go out of scope to get the output file written.
  {
    Book *book = NULL;
    int previous_book_index = -1;
    HtmlList html (htmlfilename, "Search results");
    // Produce a html file with all references in it.
    for (unsigned int i = 0; i < references.size(); i++) {
      // Close window?
      if (dialog_is_about_to_be_deleted) return;
      // Calculate the percentage complete.
      percentage = double (i) / double (references.size());
      // Get the reference decoded, and get the book number.
      string bookref, chapterref, verse;
      decode_reference (references[i], bookref, chapterref, verse);
      int book_index = scripture.get_index_of_book (bookref);
      // Book exists?
      if (book_index >= 0) {
        // Open the book.
        // To speed things up we do not destroy and recreate a book object 
        // if the book is still the same. Only new books destroy/create the object.
        // A test showed that this works more than three times as fast as when 
        // recreating the book object for each references.
        if (book_index != previous_book_index) {
          if (book) delete (book);
          book = new Book (scripture.paths[book_index], false, false);
          previous_book_index = book_index;
        }
        // Get the chapter and the verse text.
        vector <ustring> lines;
        book->get_chapter (convert_to_int (chapterref), lines);
        Chapter chapter (lines);
        vector <ustring> verses;
        chapter.get_verse (convert_to_int (verse), verses);
        // As a verse may consist of more lines, deal with each.
        ustring line;
        for (unsigned int i2 = 0; i2 < verses.size (); i2++) {
          line.append (verses[i2] + " ");
        }
        // Positions in the line, and lengths to highlight.
        vector <size_t> highlight_positions;
        vector <size_t> highlight_lengths;
        // Go through all the words to highlight.
        for (unsigned int i2 = 0; i2 < casesensitives.size(); i2++) {
          // Word to highlight
          ustring highlightword;
          if (casesensitives[i2]) highlightword = highlightwords[i2];
          else highlightword = highlightwords[i2].casefold ();
          // Variabele s holds a shadow string.
          ustring s;
          if (casesensitives[i2]) s = line;
          else s = line.casefold ();
          // Find positions for highlighting.
          size_t offposition = s.find (highlightword);
          while (offposition != string::npos) {
            // Store position and length.
            highlight_positions.push_back (offposition);
            highlight_lengths.push_back (highlightword.length());
            // Look for possible next word to highlight.
            offposition = s.find (highlightword, offposition + highlightword.length ());
          }
        }
        // Sort the positions from small to bigger.
        html_sort_positions (highlight_positions, highlight_lengths);
        // Combine overlapping positions.
        html_combine_overlaps (highlight_positions, highlight_lengths);
        // Change <, > and & to their corresponding entities.
        html_handle_entities (line, highlight_positions);
        // Insert the code for highlighting.
        html_insert_emphasis (line, highlight_positions, highlight_lengths);
        // Write the line to the html file.     
        html.writeline (references[i] + " " + line);
      }
    }
    // If object book is still alive, destroy it.
    if (book) delete (book);
  }
  
  // View the result.
  htmlbrowser (htmlfilename);

  // Log: ready.
  log_message ("Ready printing the references");
  
  // We're now done.
  percentage = 1;
  gtk_label_set_text (GTK_LABEL (label1), "The browser has the references.\nMake the browser visible to view them.");

  // Close window after a while.
  event_source_id2 = g_timeout_add_full (G_PRIORITY_DEFAULT, 5000, GSourceFunc (on_close_timeout), gpointer(this), NULL);
}


bool
PrintReferencesDialog::on_close_timeout (gpointer data)
{
  return ((PrintReferencesDialog *) data)->on_close ();
}


bool
PrintReferencesDialog::on_close ()
{
  gtk_dialog_response (GTK_DIALOG (dialogprintreferences), GTK_RESPONSE_OK);
  return false;
}
