/*
** Copyright (C) 2003 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include <gtk/gtk.h>
#include "dialogunicode.h"
#include "utilities.h"
#include "bible.h"


#define TEMPFILE "bibledit.unicode"
#define ANYTEXT "Rescue whatever text is available"


static bool dialog_is_about_to_be_deleted;


UnicodeDialog::UnicodeDialog (const string & filetoconvert)
{
  // Store variabeles
  filename = filetoconvert;
  dialog_is_about_to_be_deleted = false;
  current_encoding = 0;
  possible_encodings.clear();
  working_encodings.clear();
  
  unicodedialog = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (unicodedialog), "Convert to UTF-8 Unicode");
  gtk_window_set_position (GTK_WINDOW (unicodedialog), GTK_WIN_POS_CENTER_ON_PARENT);
  gtk_window_set_modal (GTK_WINDOW (unicodedialog), TRUE);
  gtk_window_set_default_size (GTK_WINDOW (unicodedialog), 480, 480);
  // Set the type to a splash screen, so that the window manager does not show
  // a close button. Ideally the delete_event handler should prevent the window
  // from closing, but for one or another reason this does not work.
  // This splash screen hack therefore, is needed to prevent the user from closing
  // the window.
  gtk_window_set_type_hint (GTK_WINDOW (unicodedialog), GDK_WINDOW_TYPE_HINT_DIALOG);

  dialog_vbox1 = GTK_DIALOG (unicodedialog)->vbox;
  gtk_widget_show (dialog_vbox1);

  vbox1 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox1);
  gtk_box_pack_start (GTK_BOX (dialog_vbox1), vbox1, TRUE, TRUE, 0);

  // First informative label
  label1 = gtk_label_new (
    "The book you are now opening is not in the right Unicode format.\n"
    "It needs to be converted before use.\n\n"
    "Looking up possible ways of converting it.\n"
    "Please wait ...");
  gtk_widget_show (label1);
  gtk_box_pack_start (GTK_BOX (vbox1), label1, FALSE, FALSE, 4);
  gtk_misc_set_alignment (GTK_MISC (label1), 0, 0.5);

  combo1 = gtk_combo_new ();
  gtk_widget_show (combo1);
  gtk_box_pack_start (GTK_BOX (vbox1), combo1, FALSE, FALSE, 9);

  combo_entry1 = GTK_COMBO (combo1)->entry;
  gtk_widget_show (combo_entry1);
  gtk_editable_set_editable (GTK_EDITABLE (combo_entry1), FALSE);
  gtk_entry_set_activates_default (GTK_ENTRY (combo_entry1), TRUE);

  progressbar1 = gtk_progress_bar_new ();
  gtk_widget_show (progressbar1);
  gtk_box_pack_start (GTK_BOX (vbox1), progressbar1, FALSE, FALSE, 7);

  // Progress bar.
  scrolledwindow1 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow1);
  gtk_box_pack_start (GTK_BOX (vbox1), scrolledwindow1, TRUE, TRUE, 2);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow1), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  textview1 = gtk_text_view_new ();
  gtk_widget_show (textview1);
  gtk_container_add (GTK_CONTAINER (scrolledwindow1), textview1);
  gtk_text_view_set_editable (GTK_TEXT_VIEW (textview1), FALSE);
  gtk_text_view_set_pixels_below_lines (GTK_TEXT_VIEW (textview1), 2);
  gtk_text_view_set_pixels_inside_wrap (GTK_TEXT_VIEW (textview1), 2);

  dialog_action_area1 = GTK_DIALOG (unicodedialog)->action_area;
  gtk_widget_show (dialog_action_area1);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area1), GTK_BUTTONBOX_END);

  okbutton1 = gtk_button_new_from_stock ("gtk-ok");
  gtk_widget_show (okbutton1);
  gtk_dialog_add_action_widget (GTK_DIALOG (unicodedialog), okbutton1, GTK_RESPONSE_OK);
  gtk_widget_set_sensitive (okbutton1, FALSE);
  GTK_WIDGET_SET_FLAGS (okbutton1, GTK_CAN_DEFAULT);

  g_signal_connect ((gpointer) combo_entry1, "changed",
                    G_CALLBACK (unicodedialog_on_combo_entry1_changed),
                    gpointer(this));
  g_signal_connect ((gpointer) okbutton1, "clicked",
                    G_CALLBACK (unicodedialog_on_okbutton1_clicked),
                    gpointer(this));
  g_signal_connect (G_OBJECT (unicodedialog), "delete_event",
		            G_CALLBACK (delete_event), 
		            gpointer(this));

  gtk_widget_grab_focus (combo_entry1);
  gtk_widget_grab_default (okbutton1);

  // Get the whole list of possible encodings.
  FILE *stream;
  string s = "iconv --list";
  stream = popen (s.c_str (), "r");
  char buf[1024];
  while (fgets (buf, sizeof (buf), stream))
    {
      s = buf;
      s = trim (s);
      if (g_str_has_suffix (s.c_str(), "//"))
        s = s.substr (0, s.length () - 2);
      possible_encodings.push_back (s);
    }
  pclose (stream);

  // Filter out the encodings that would produce results.
  event_source_id = g_timeout_add_full (G_PRIORITY_DEFAULT, 20, GSourceFunc (unicodedialog_filter_encodings), gpointer(this), NULL);
  g_thread_create (GThreadFunc (unicodedialog_thread_filter_encodings), gpointer(this), false, NULL);
}


UnicodeDialog::~UnicodeDialog ()
{
  unlink (temporary_file (TEMPFILE).c_str ());
  gtk_widget_destroy (unicodedialog);
}


int UnicodeDialog::run ()
{
  return gtk_dialog_run (GTK_DIALOG (unicodedialog));
}


void
UnicodeDialog::on_ok ()
{
  /*
     Deal with response OK: copy the converted file.
   */
  string command;
  command = "cp --force ";
  command.append (temporary_file (TEMPFILE));
  command.append (" '");
  command.append (filename);
  command.append ("'");
  system (command.c_str());
}


bool
UnicodeDialog::filter_encodings ()
{
  /*
   * This function manages the display while the thread looks up the encodings 
   * that actually would be able to convert the input file.
   * Then it prepares the next screen.
   */
  // Set the progress bar.
  double fraction = double (current_encoding) / possible_encodings.size ();
  // Sometimes the fraction gets too high and crashes the program. Limit it.
  if (fraction > 1)
    fraction = 1;
  gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (progressbar1), fraction);
  // If all encodings have been dealt with, clean up, and present the next screen.
  if (current_encoding >= possible_encodings.size ())
    {
      // Clear some unused memory.
      possible_encodings.clear ();
      // Change the screen a bit.
      gtk_widget_destroy (progressbar1);
      gtk_label_set_text (GTK_LABEL (label1),
        "Here is a list of possible encodings\n"
        "If one is selected, it will show the result as well\n"
        "When you're happy with the result, press OK.\n"
        "The book will then be converted.");
      working_encodings.push_back (ANYTEXT);

      // Load the entries that work into the combobox.
      GList *combo1_items = NULL;
      for (unsigned int i = 0; i < working_encodings.size(); i++)
        combo1_items = g_list_append (combo1_items, (gpointer) working_encodings[i].c_str());
      gtk_combo_set_popdown_strings (GTK_COMBO (combo1), combo1_items);
      g_list_free (combo1_items);
      // Display the first entry.
      gtk_entry_set_text (GTK_ENTRY (combo_entry1), working_encodings[0].c_str());
      on_combo_changed ();
      // Allow window to close now.
      gtk_widget_set_sensitive (okbutton1, TRUE);
      return false;
    }
  else
    {
      return true;
    }
}


void
UnicodeDialog::thread_filter_encodings ()
{
  /*
     This thread looks up the encodings that actually would be able to convert the input file.
     Conversions that don't work are removed.
   */

  while ((current_encoding < possible_encodings.size ()))
    {
      if (dialog_is_about_to_be_deleted)
      	return;
      // The command to find out about the current encoding, whether it would work.
      string command;
      command = "iconv -f '";
      command.append (possible_encodings[current_encoding]);
      command.append ("' -t UTF-8 '");
      command.append (filename);
      command.append ("' > /dev/null 2>&1");
      int shellresult = system (command.c_str());
      if (dialog_is_about_to_be_deleted)
      	return;
      // If conversion succeeded, store this encoding as a possible one.
      if (shellresult == 0)
        {
          working_encodings.push_back (possible_encodings[current_encoding]);
        }
      // Next encoding.
      current_encoding++;
    }
}


void
UnicodeDialog::on_combo_changed ()
{
  /*
     If the user selects an encoding, this will show the example as well.
   */
  // Convert the text to a temporary file using the selected encoding.
  string s = gtk_entry_get_text (GTK_ENTRY (combo_entry1));
  if (s.empty ())
    return;
  // See whether we'll use iconv or strings to filter the input text.
  string command;
  if (s == ANYTEXT)
    {
      command = "strings --bytes=2";
    }
  else
    {
      command = "iconv -f ";
      command.append (s);
      command.append (" -t UTF-8");
    }
  command.append (" '");
  command.append (filename);
  command.append ("' > ");
  command.append (temporary_file (TEMPFILE));
  system (command.c_str ());
  // Display the result.
  gchar *contents;
  g_file_get_contents (temporary_file (TEMPFILE).c_str(), &contents, NULL, NULL);
  gtk_text_buffer_set_text (gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview1)), contents, -1);
  g_free (contents);
}


void UnicodeDialog::unicodedialog_on_combo_entry1_changed  (GtkEditable     *editable,
                                        gpointer         user_data)
{
  ((UnicodeDialog *) user_data)->on_combo_changed();
}


void UnicodeDialog::unicodedialog_on_okbutton1_clicked     (GtkButton       *button,
                                        gpointer         user_data)
{
  ((UnicodeDialog *) user_data)->on_ok();
}


bool UnicodeDialog::unicodedialog_filter_encodings (gpointer data)
{
  return ((UnicodeDialog *) data)->filter_encodings();
}


void UnicodeDialog::unicodedialog_thread_filter_encodings (gpointer data)
{
  ((UnicodeDialog *) data)->thread_filter_encodings();
}


gboolean UnicodeDialog::delete_event(GtkWidget *widget, GdkEvent  *event, gpointer   data )
{
  /* If you return FALSE in the "delete_event" signal handler,
   * GTK will emit the "destroy" signal. Returning TRUE means
   * you don't want the window to be destroyed.
   * This is useful for popping up 'are you sure you want to quit?'
   * type dialogs. */
  // Set flag we're about to get destroyed.
  dialog_is_about_to_be_deleted = true;
  // Destroy the event source, if it is still alive, 
  // so that the timeout function is no longer called.
  GSource *source;
  source = g_main_context_find_source_by_id (NULL, ((UnicodeDialog *) data)->event_source_id);
  if (source)
    g_source_destroy (source);
  // Wait a while to let the function that drives the display finish.
  g_usleep (40000);
  /* Change TRUE to FALSE and the main window will be destroyed with
   * a "delete_event". */
  /* So says the documentation. It does work in a main window, but not in
   * dialogs that pop up. I've filed bug 
   * http://bugzilla.gnome.org/show_bug.cgi?id=167187
   * Once the bug has been fixed, and the minimal version of Gtk+ is equal to
   * or greater than the fixed version, we can change the work-around.
   */
  return TRUE;
}
