/*
** Copyright (C) 2003 The Free Software Foundation.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include "utilities.h"
#include "html_book.h"
#include <config.h>
#include "usfmtools.h"
#include "constants.h"


HtmlBook::HtmlBook(const string& filename, const string& runningheader, 
                   const string& previoustitle, const string& previouslink, 
                   const string& nexttitle, const string& nextlink,
                   const string& indexlink, const vector<string>& headers)
: writetext (filename),
  usfm (0)
{
  // Store/initialize data to be used later.
  myprevioustitle = previoustitle;
  mypreviouslink = previouslink;
  mynexttitle = nexttitle;
  mynextlink = nextlink;
  myindexlink = indexlink;
  paragraph_opened = false;
  // Header
  writetext.out << "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">" << endl;
  writetext.out << "<html>" << endl;
  writetext.out << "<head>" << endl;
  writetext.out << "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\">" << endl;
  writetext.out << "<title>" << bold_strikeout (runningheader) << "</title>" << endl;
  writetext.out << "<meta name=\"generator\" content=\"Bibledit\">" << endl;
  writetext.out << "</head>" << endl;
  writetext.out << "<body bgcolor=\"white\" text=\"black\" link=\"#0000FF\" vlink=\"#840084\" alink=\"#0000FF\">" << endl;
  writetext.out << "<table width=\"100%\">" << endl;
  writetext.out << "<tr>" << endl;
  writetext.out << "<th colspan=\"3\" align=\"center\">" << bold_strikeout (runningheader) << "</th>" << endl;
  writetext.out << "</tr>" << endl;
  writetext.out << "<tr>" << endl;
  writetext.out << "<td width=\"20%\" align=\"left\">";
  if (previouslink.length() > 0)
    writetext.out << "<a href=\"" << previouslink << "\">Prev</a>";
  writetext.out << "</td>" << endl;
  writetext.out << "<th width=\"60%\" align=\"center\"></th>" << endl;
  writetext.out << "<td width=\"20%\" align=\"right\">";
  if (nextlink.length() > 0)
    writetext.out << "<a href=\"" << nextlink << "\">Next</a>";
  writetext.out << "</td>" << endl;
  writetext.out << "</tr>" << endl;
  writetext.out << "</table>" << endl;
  writetext.out << "<hr>" << endl;
  if (headers.size() > 0) {
    Usfm usfm(0);
    for (unsigned int i = 0; i < headers.size(); i++) {
      ustring line (headers[i]);
      ustring marker = usfm_extract_marker (line);
      if (usfm.is_maintitle(marker))
        writetext.out << "<h1>" << bold_strikeout (line) << "</h1>" << endl;
      if (usfm.is_subtitle(marker))
        writetext.out << "<h2>" << bold_strikeout (line) << "</h2>" << endl;
    }
  } else {
    writetext.out << "<h1>" << bold_strikeout (runningheader) << "</h1>" << endl;
  }
  writetext.out << "<hr>" << endl;
}


HtmlBook::~HtmlBook()
{
  // Close the html file.
  writetext.out << "<hr>" << endl;
  writetext.out << "<table width=\"100%\">" << endl;
  writetext.out << "<tr>" << endl;
  writetext.out << "<td width=\"40%\" align=\"left\">";
  if (mypreviouslink.length() > 0)
    writetext.out << "<a href=\"" << mypreviouslink << "\">Prev</a></td>" << endl;
  writetext.out << "<td width=\"20%\" align=\"center\"></td>" << endl;
  writetext.out << "<td width=\"40%\" align=\"right\">";
  if (mynextlink.length() > 0)
    writetext.out << "<a href=\"" << mynextlink << "\">Next</a>";
  writetext.out << "</td>" << endl;
  writetext.out << "</tr>" << endl;
  writetext.out << "<tr>" << endl;
  writetext.out << "<td width=\"40%\" align=\"left\" valign=\"top\">" << bold_strikeout (myprevioustitle) << "</td>" << endl;
  writetext.out << "<td width=\"20%\" align=\"center\">";
  if (myprevioustitle.length() > 0)
    writetext.out << "<a href=\"" << myindexlink << "\">Home</a></td>" << endl;
  writetext.out << "<td width=\"40%\" align=\"right\" valign=\"top\">" << bold_strikeout (mynexttitle) << "</td>" << endl;
  writetext.out << "</tr>" << endl;
  writetext.out << "</table>" << endl;
  writetext.out << "</body>" << endl;
  writetext.out << "</html>" << endl;
}


void HtmlBook::openindex ()
{
  writetext.out << "<p><b>Table of Contents</b></p>" << endl;
  writetext.out << "<dl>" << endl;
}


void HtmlBook::closeindex ()
{
  writetext.out << "</dl>" << endl;
  writetext.out << "<p>Produced by Bibledit</p>" << endl;
  time_t value = time (0);
  string current_date = ctime (&value);
  current_date = trim (current_date);
  writetext.out << "<p>" << endl;
  writetext.out << current_date << endl;
  writetext.out << "</p>" << endl;
}


void HtmlBook::writelink (const string& title, const string& link)
{
  writetext.out << "<dt><a href=\"" << link << "\">" << bold_strikeout(title) << "</a></dt>" << endl;
}


void HtmlBook::openchapter (unsigned int chapternumber, const string& runningheader)
// Writes the chapter number.
// For clarity, when the html pages are printed or so, the name of the book will be added too.
{
  writetext.out << "<h2><a name=\"" << chapternumber << "\"></a>" << bold_strikeout(runningheader) << " " << chapternumber << "</h2>" << endl;
}


void HtmlBook::closechapter ()
{
}


void HtmlBook::openparagraph ()
{
  if (paragraph_opened) 
    closeparagraph();
  writetext.out << "<p style=\"text-align: justify;\">" << endl;
  paragraph_opened = true;
}


void HtmlBook::closeparagraph ()
{
  if (paragraph_opened)
    writetext.out << "</p>" << endl;
  paragraph_opened = false;
}


void HtmlBook::writeline (const string& line)
{
  writetext.out << bold_strikeout (line) << endl;
}


ustring HtmlBook::bold_strikeout (const ustring& line)
// Deal with the bold and strikeout markers.
// The line may have bold and strike-out markers in it.
// Change them to the proper code.
{
  ustring s (line);
  unsigned int offposition;
  unsigned int len;
  // Replace strike-through begin marker.
  len = strlen (STRIKE_THROUGH_BEGIN);
  offposition = s.find (STRIKE_THROUGH_BEGIN);
  while (offposition != string::npos) {
    s.replace (offposition, len, "<strike>");
    offposition = s.find (STRIKE_THROUGH_BEGIN, offposition);
  }
  // Replace strike-through end marker.
  len = strlen (STRIKE_THROUGH_END);
  offposition = s.find (STRIKE_THROUGH_END);
  while (offposition != string::npos) {
    s.replace (offposition, len, "</strike>");
    offposition = s.find (STRIKE_THROUGH_END, offposition);
  }
  // Replace bold begin marker.
  len = strlen (BOLD_BEGIN);
  offposition = s.find (BOLD_BEGIN);
  while (offposition != string::npos) {
    s.replace (offposition, len, "<b>");
    offposition = s.find (BOLD_BEGIN, offposition);
  }
  // Replace bold end marker.
  len = strlen (BOLD_END);
  offposition = s.find (BOLD_END);
  while (offposition != string::npos) {
    s.replace (offposition, len, "</b>");
    offposition = s.find (BOLD_END, offposition);
  }
  return s;
}
