/*
  ** Copyright (C) 2003 The Free Software Foundation.
  **  
  ** This program is free software; you can redistribute it and/or modify
  ** it under the terms of the GNU General Public License as published by
  ** the Free Software Foundation; either version 2 of the License, or
  ** (at your option) any later version.
  **  
  ** This program is distributed in the hope that it will be useful,
  ** but WITHOUT ANY WARRANTY; without even the implied warranty of
  ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  ** GNU General Public License for more details.
  **  
  ** You should have received a copy of the GNU General Public License
  ** along with this program; if not, write to the Free Software
  ** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  **  
  */


#include "libraries.h"
#include "utilities.h"
#include <libgen.h>
#include <glib.h>
#include <config.h>
#include "notes_utils.h"
#include "directories.h"
#include "sqlite3_plus.h"
#include "bible.h"
#include "date_time_utils.h"


void notes_database_verify_create ()
/*
This verifies the notes database is fine.
If it's not there, it creates one.
*/
{
  try
  {
    // Connect to the database
    sqlite3::connection con (notes_database_filename ().c_str ());
    int count;
    // Check for the table
    count = con.executeint32("select count(*) from sqlite_master where name='%s';", TABLE_NOTES);
    if (count == 0)
      // Create the table
      con.executenonquery("create table '%s' (id integer, reference text, project text, status integer, referred integer, note text, casefolded text, created integer, modified integer, user text);", TABLE_NOTES);
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Initializing notes database: " << ex.what () << endl;
  }
}


string notes_database_filename ()
// Returns the filename of the notes database.
{
  return g_build_filename (directories_get_notes ().c_str(), "database", NULL);
}


void insert_link (GtkTextBuffer *buffer, string text, gint id)
/* Inserts a piece of text into the buffer, giving it the usual
 * appearance of a hyperlink in a web browser: blue and underlined.
 * Additionally, attaches some data on the tag, to make it recognizable
 * as a link. 
 */
{
  GtkTextTag *tag;
  tag = gtk_text_buffer_create_tag (buffer, NULL, 
				    "foreground", "blue", 
				    "underline", PANGO_UNDERLINE_SINGLE, 
				    NULL);
  g_object_set_data (G_OBJECT (tag), "id", GINT_TO_POINTER (id));
  GtkTextIter iter;
  gtk_text_buffer_get_end_iter (buffer, &iter);
  gtk_text_buffer_insert_with_tags (buffer, &iter, text.c_str(), -1, tag, NULL);
}
  

gint notes_database_get_unique_id ()
// This generates a unique id, one that is not yet used in the notes database.
{
  gint32 result;
  try
  {
    sqlite3::connection con (notes_database_filename ().c_str ());
    bool found = false;
    while (!found) {
      result = g_random_int_range (1, 100000000);
      gint count = con.executeint32("select count(*) from '%s' where id=%i;", TABLE_NOTES, result);
      if (count == 0) {
        found = true;
      }
    }
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Getting random id for notes database: " << ex.what () << endl;
  }
  return result;
}


string notes_index_to_referred_to (int index)
{
  if (index == 0) return "Nobody";
  if (index == 1) return "Myself";
  if (index == 2) return "Subteam";
  if (index == 3) return "Team";
  if (index == 4) return "Panel";
  if (index == 5) return "Church";
  if (index == 6) return "Consultant";
  if (index == 7) return "Bible Society";
  return "";
}


int notes_referred_to_index (const string& referred)
{
  for (int i = 0; i < NOTES_NUMBER_OF_REFERRED_TO_S; i++)
    if (referred == notes_index_to_referred_to(i))
      return i;
  return 0;
}


void notes_delete_one (int id)
// Deletes the note with id.
{
  try
  {
    // Connect to database and start transaction
    sqlite3::connection con (notes_database_filename ().c_str ());
    con.executenonquery ("begin;");
    // Delete data with "id".
    con.executenonquery ("delete from %s where id = %d;", TABLE_NOTES, id);
    // Commit the transaction and close connection.
    con.executenonquery ("commit;");
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Deleting note: " << ex.what () << endl;
  }
}


void notes_sort (vector<gint32>& ids, const vector<string>& refs, const vector<int>& dates)
/*
This sorts notes.
ids         - ID's of the notes.
              The ids will be modified and sorted, so that these can be used 
              to fetch data from the database in the right order.
refs        - References to sort on first.
dates       - Modification dates to sort on next.
*/
{
  // No sorting when less than two entries.
  if (ids.size() < 2)
    return;

  // The sorting is done by putting everything in one string, and then
  // sorting the vector of strings, and then taking out the ids again.

  // Storage for strings to sort
  vector<string> strings_to_sort;
  
  // Go through all data to sort.
  for (unsigned int i = 0; i < ids.size(); i++) {
  
    // Storage for string to build.
    string string_to_build;
    
    // First part of the string is the reference.
    // It always has the same length, so can be taken as it.
    string_to_build = refs[i];

    // Next part to sort on is the date. This Julian day has a length of 6
    // characters at the time of programming, and will continue so until
    // Deo Volente the year 2739 A.D. Therefore there is no need to modify the
    // length of it.
    {
      string date;
      date = convert_to_string (dates[i]);
      string_to_build.append(date);
    }
    
    // Last part of the string to sort is the id.
    // Maximum id = 100000000, so maximum length of 9 characters.
    {
      string id;
      id = "000000000" + convert_to_string (ids[i]);
      id = id.substr(id.length() - 9, 9);
      string_to_build.append (id);
    }

    // Store built string
    strings_to_sort.push_back(string_to_build);    
  }

  // Sort the numerical equivalents.
  sort (strings_to_sort.begin(), strings_to_sort.end());
  
  // Extract the sorted ids from the sorted data.
  ids.clear();
  for (unsigned int i = 0; i < strings_to_sort.size(); i++) {
    string id = strings_to_sort[i].substr(15, 9);
    gint32 numerical_id = convert_to_int (id);
    ids.push_back(numerical_id);
  }
}


void notes_select (NotesViewSelector selector, vector<gint32>& ids, const string currentreference)
/*
This selects notes for display.
It does this based on criteria such as
- selector - which selection to make.
- currentreference - if we show notes for the current verse only, this shows which verse that is.
The resulting selection will be stored in ids.
*/
{
  // Clear ids.
  ids.clear();
  try
  {
    // Connect to database.
    sqlite3::connection con (notes_database_filename ().c_str ());
    // First read from database, for sorting purposes.
    sqlite3::reader reader;
    // See which notes to select.
    switch (selector) {
      case CurrentVerse:
      {
        string book, chapter, verse;
        decode_reference (currentreference, book, chapter, verse);
        string this_verse;
        this_verse = reference_to_numerical_equivalent (book, chapter, verse);      
        reader = con.executereader ("select id, reference, modified from '%s' where reference glob ('*%s*');", TABLE_NOTES, this_verse.c_str());
        break;
      }
      case AllNotes:
      {
        reader = con.executereader ("select id, reference, modified from '%s';", TABLE_NOTES);
        break;
      }
      case TodaysNotes:
      {
        int currentdate = date_time_julian_day_get_current ();
        reader = con.executereader ("select id, reference, modified from '%s' where modified = %d;", TABLE_NOTES, currentdate);
        break;
      }
      case MyNotes:
      {
        // The "1" refers to "referred to MyNotes".
        reader = con.executereader ("select id, reference, modified from '%s' where referred = %d;", TABLE_NOTES, 1);
        break;
      }
      case SubteamNotes:
      {
        // The "2" refers to "referred to Subteam".
        reader = con.executereader ("select id, reference, modified from '%s' where referred = %d;", TABLE_NOTES, 2);
        break;
      }
      case TeamNotes:
      {
        // The "3" refers to "referred to Team".
        reader = con.executereader ("select id, reference, modified from '%s' where referred = %d;", TABLE_NOTES, 3);
        break;
      }
      case PanelNotes:
      {
        reader = con.executereader ("select id, reference, modified from '%s' where referred = %d;", TABLE_NOTES, 4);
        break;
      }
      case ChurchNotes:
      {
        reader = con.executereader ("select id, reference, modified from '%s' where referred = %d;", TABLE_NOTES, 5);
        break;
      }
      case ConsultantNotes:
      {
        reader = con.executereader ("select id, reference, modified from '%s' where referred = %d;", TABLE_NOTES, 6);
        break;
      }
      case BibleSocietyNotes:
      {
        reader = con.executereader ("select id, reference, modified from '%s' where referred = %d;", TABLE_NOTES, 7);
        break;
      }
    }
    // Storage for sorting purposes.
    vector<string> references;
    vector<int> dates;
    // Read all resulting data from the db.
    while (reader.read ()) {
      // Get id.
      gint32 id = reader.getint32(0);
      // Get the numerical equivalent of the reference.
      string reference;
      reference = reader.getstring (1);
      {
        // Parse the string into its possible several references.
        Parse parse (reference);
        // Take the first refence available.
        if (parse.words.size() > 0)
          reference = parse.words[0];
      }
      // Get date modified.
      int date;
      date = reader.getint32(2);
      // Store data.
      ids.push_back(id);
      references.push_back(reference);
      dates.push_back(date);
    }
    // Close database
    reader.close ();
    con.close ();
    // Sort the notes.
    notes_sort (ids, references, dates);
  }
  catch (exception & ex)
  {
    cerr << "Error: " << ex.what () << endl;
  }
}


void notes_display_one (GtkWidget *textview, GtkTextBuffer *textbuffer, gint32 id)
/*
This cares for the actual displaying of the notes.
It displays the note with an ID of id.
*/
{
  try
  {
    // Connect to database.
    sqlite3::connection con (notes_database_filename ().c_str ());
    // We've got an id display here. 
    
    // Get a reader for the database.
    sqlite3::reader reader;
    // Execute the SQL statement.
    reader = con.executereader ("select id, reference, note from '%s' where id = %d;", TABLE_NOTES, id);
    // Go through the results (there should be only one anyway).
    while (reader.read ()) {

      // Get the numerical equivalent of the reference.
      string reference;
      reference = reader.getstring (1);

      // Parse the string into its possible several references.
      Parse parse (reference);
      reference.clear();

      // Go through each reference.
      for (unsigned int i2 = 0; i2 < parse.words.size(); i2++) {
        // Make it human readable.
        int book, chapter, verse;
        numerical_equivalent_to_reference (parse.words[i2], book, chapter, verse);
        if (!reference.empty())
          reference.append (", ");
        reference.append (book_chapter_verse_to_reference (book, chapter, verse));
      }

      // Insert a link with this heading.
      insert_link (textbuffer, reference, id); 
      gtk_text_buffer_insert_at_cursor (textbuffer, "\n", -1);
        
      // Get the text of the note.
      ustring note;
      note = reader.getstring (2);
      gtk_text_buffer_insert_at_cursor (textbuffer, note.c_str(), -1);
      gtk_text_buffer_insert_at_cursor (textbuffer, "\n", -1);
        
    }
    // Close reader.
    reader.close ();
      
    // Close connection
    con.close ();
  }
  catch (exception & ex)
  {
    cerr << "Error: " << ex.what () << endl;
  }
  // We're through, so clear the ids to display to save memory.
}


void notes_get_references_from_editor (GtkTextBuffer *textbuffer, vector<string>& references, vector<string>& messages)
/*
Gets all references from the notes editor.
Normalizes them.
Produces messages on trouble.
*/
{
  // Get all lines from the textbuffer.
  vector<ustring> lines;
  textbuffer_get_lines (textbuffer, lines);
  // When discovering a reference from a user's entry, use previous values,
  // so that it becomes quicker for a user to enter new references.
  // If Leviticus 10:11 is already there, and the user wishes to add verse
  // 12 also, he just enters 12 on a line, and that' it.
  string previousbook, previouschapter, previousverse;
  for (unsigned int i = 0; i < lines.size(); i++) {
    if (!lines[i].empty ()) {
      // Normalize reference.
      string book;
      string chapter;
      string verse;
      if (reference_discover (previousbook, previouschapter, previousverse, lines[i], book, chapter, verse)) {
        // Store reference.
        string ref = book + " " + chapter + ":" + verse;
        references.push_back (ref);
        // Store values to discover next reference.
        previousbook = book;
        previouschapter = chapter;
        previousverse = verse;
      } else {
        messages.push_back ("Reference " + lines[i] + " is not valid and has been removed.");
      }
    }
  }
}
