/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "libraries.h"
#include "utilities.h"
#include "references.h"
#include "directories.h"
#include "constants.h"
#include "referenceutils.h"
#include "bible.h"


References::References (GtkListStore * liststore, GtkWidget * treeview, GtkTreeViewColumn * treecolumn)
{
  myliststore = liststore;
  mytreeview = treeview;
  mytreecolumn = treecolumn;
}


References::~References ()
{
}


void
References::load (const string & filename)
{
  // Loads references from a file.
  // If no filename is given, the default name is used.
  string file (filename);
  if (file.length () == 0)
    file = references_default_filename ();
  try
  {
    ReadText rt (file, true);
    // Pick out the references and leave the rest.
    string book;
    string chapter;
    string verse;
    for (unsigned int i = 0; i < rt.lines.size (); i++)
      if (reference_discover ("", "", "", rt.lines[i], book, chapter, verse))
        references.push_back (book + " " + chapter + ":" + verse);
    sort_references (references);
    load_searchwords (file);
  }
  catch (exception & ex)
  {
    cerr << "Loading references: " << ex.what () << endl;
  }
}


void
References::fill_store ()
{
  /*
   * Fill the list store with the data.
   */
  // Clear the store first.
  gtk_list_store_clear (myliststore);
  // Now add new data.
  GtkTreeIter iter;
  for (unsigned int i = 0; i < references.size (); i++)
    {
      gtk_list_store_append (myliststore, &iter);
      gtk_list_store_set (myliststore, &iter, 0, references[i].c_str (), -1);
    }
  // Save references to default location.
  save ("");
  // Show number of references in the title of the column
  set_header ();
}


void
References::set_header ()
{
  // Sets the header in the treeview to the number of references.
  string s = TEXT_REFERENCES;
  s.append (" - ");
  s.append (convert_to_string (int (references.size ())));
  gtk_tree_view_column_set_title (mytreecolumn, s.c_str ());
}


void
References::save (const string & filename)
{
  // Saves references to a file.
  // If no filename is given, the default name is used.
  string file (filename);
  if (file.length () == 0)
    file = references_default_filename ();
  try
  {
    write_lines (file, references);
  }
  catch (exception & ex)
  {
    cerr << "Saving references: " << ex.what () << endl;
  }
  save_searchwords (file);
}


void
References::goto_next ()
{
  /* 
   * This selects the next reference, if there is any.
   * If no item has been selected it selects the first, if it's there.
   */
  goto_next_previous_internal (true);
}


void
References::goto_previous ()
{
  /* 
   * This goes to the previous reference, if there is any.
   * If no item has been selected it chooses the first, if it's there.
   */
  goto_next_previous_internal (false);
}


void
References::goto_next_previous_internal (bool next)
{
  // Continue only when references are available.
  if (gtk_tree_model_get_iter_first (GTK_TREE_MODEL (myliststore), &iterator))
    {
      // See which is selected.
      GtkTreeSelection *selection;
      selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (mytreeview));
      // See how many are selected.
      gint number_selected;
      number_selected = gtk_tree_selection_count_selected_rows (selection);
      if (number_selected > 0)
        {
          // One or more selected: Store pointer to last one selected, and select next.
          gtk_tree_selection_selected_foreach (selection, References::goto_foreach_function, gpointer (&iterator));
          if (next)
            {
              if (gtk_tree_model_iter_next (GTK_TREE_MODEL (myliststore), &iterator))
                {
                  gtk_tree_selection_unselect_all (selection);
                  gtk_tree_selection_select_iter (selection, &iterator);
                }
            }
          else
            {
              GtkTreePath *path;
              path = gtk_tree_model_get_path (GTK_TREE_MODEL (myliststore), &iterator);
              if (gtk_tree_path_prev (path))
                {
                  gtk_tree_selection_unselect_all (selection);
                  gtk_tree_model_get_iter (GTK_TREE_MODEL (myliststore), &iterator, path);
                  gtk_tree_selection_select_iter (selection, &iterator);
                }
            }
        }
      else
        {
          // None selected: select the first one.
          gtk_tree_selection_select_iter (selection, &iterator);
        }
      // Scroll, if needed, to make the new selection visible.
      if (gtk_list_store_iter_is_valid (myliststore, &iterator))
        {
          GtkTreePath *path;
          path = gtk_tree_model_get_path (GTK_TREE_MODEL (myliststore), &iterator);
          if (next)
            gtk_tree_view_scroll_to_cell (GTK_TREE_VIEW (mytreeview), path, NULL, true, 0.9, 0);
          else
            gtk_tree_view_scroll_to_cell (GTK_TREE_VIEW (mytreeview), path, NULL, true, 0.1, 0);
          gtk_tree_path_free (path);
        }
    }
}


void
References::goto_foreach_function (GtkTreeModel * model, GtkTreePath * path, GtkTreeIter * iter, gpointer data)
{
  GtkTreeIter *iterator;
  iterator = (GtkTreeIter *) data;
  *iterator = *iter;
}


void
References::get_loaded ()
{
  // This gets the references that are loaded in the editor.
  references.clear ();
  gtk_tree_model_foreach (GTK_TREE_MODEL (myliststore), GtkTreeModelForeachFunc (loaded_get_foreach_function), gpointer (&references));
}


gboolean References::loaded_get_foreach_function (GtkTreeModel * model, GtkTreePath * path, GtkTreeIter * iter, gpointer data)
{
  gchar *
    reference;
  gtk_tree_model_get (model, iter, 0, &reference, -1);
  ((vector < string > *)data)->push_back (reference);
  g_free (reference);
  return false;
}


void
References::load_searchwords (const string & filename)
/*
  Read the file with the search words, and pick out those
  that belong to variabele filename. See the save method
  for the format of the file
*/
{
  try
  {
    searchwords.clear ();
    casesensitives.clear ();
    bool take_this = false;
    int take_what = 0;
    vector < ustring > lines;
    ReadText rt (references_searchwords_filename (), true);
    for (unsigned int i = 0; i < rt.lines.size (); i++)
      {
        ustring s = trim (rt.lines[i]);
        string first;
        string last;
        if (s.length () > 0)
          {
            first = s[0];
            last = s[s.length () - 1];
          }
        if ((first == "[") && (last == "]"))
          take_this = false;
        if (s.find (filename) != string::npos)
          take_this = true;
        if (take_this)
          {
            /*
               These are the values the take_what variabele will get:
               1: We've encountered the line containing the filename.
               2: We've encountered a searchword.
               3: We've encountered a case sensitivity
               Optionally:
               4: Searchword.
               5: Case
               6: Searchword.
               7: Case.
               Etc.
             */
            take_what++;
          }
        else
          {
            take_what = 0;
          }
        if (take_what > 1)
          {
            if ((take_what % 2) == 0)
              searchwords.push_back (s);
            if ((take_what % 2) == 1)
              casesensitives.push_back (convert_to_bool (s));
          }
      }
  }
  catch (exception & ex)
  {
    cerr << "Reading searchwords: " << ex.what () << endl;
  }
}


void
References::save_searchwords (const string & filename)
{
/*
    Whenever any file with references is saved,
    the search- and/or replacewords, that formed this 
    list of references, are saved too. Also saved is 
    whether this search and/or replace was casesensitive.
    There is one file where all data is saved in.
    A typical file might look like this:
    
    [/home/user/references]
    Word
    0
    [/home/user/bibledit/refs]
    Word1
    1
    word2
    0
    word3
    1
    
    First the name of the file with references gets stored,
    between square brackets.
    Next each search- and/or replaceword is stored, together with
    a 0 or 1 indicating whether the search/replace was case sensitive.
    Each filename can be followed by zero or more entries consisting of
    word and case sensitivity.
*/
  try
  {
    bool delete_this = false;
    vector < ustring > lines;
    ReadText rt (references_searchwords_filename (), true);
    for (unsigned int i = 0; i < rt.lines.size (); i++)
      {
        ustring s = trim (rt.lines[i]);
        string first;
        string last;
        if (s.length () > 0)
          {
            first = s[0];
            last = s[s.length () - 1];
          }
        if ((first == "[") && (last == "]"))
          delete_this = false;
        if (s.find (filename) != string::npos)
          delete_this = true;
        if (!delete_this)
          lines.push_back (s);
      }
    lines.push_back ("[" + filename + "]");
    for (unsigned int i = 0; i < searchwords.size (); i++)
      {
        lines.push_back (searchwords[i]);
        lines.push_back (convert_to_string (casesensitives[i]));
      }
    write_lines (references_searchwords_filename (), lines);
  }
  catch (exception & ex)
  {
    cerr << "Saving searchwords: " << ex.what () << endl;
  }
}


void
References::get_references (vector < string > &refs, vector < ustring > &words, vector < bool > &cases)
{
  refs.assign (references.begin (), references.end ());
  words.assign (searchwords.begin (), searchwords.end ());
  cases.assign (casesensitives.begin (), casesensitives.end ());
}


void
References::get_references (vector < string > &refs)
{
  refs.assign (references.begin (), references.end ());
}


void
References::get_references (vector < ustring > &words, vector < bool > &cases)
{
  words.assign (searchwords.begin (), searchwords.end ());
  cases.assign (casesensitives.begin (), casesensitives.end ());
}


void
References::set_references (vector < string > &refs, vector < ustring > &words, vector < bool > &cases)
{
  references.assign (refs.begin (), refs.end ());
  searchwords.assign (words.begin (), words.end ());
  casesensitives.assign (cases.begin (), cases.end ());
}
